/*
 * QLogic ISP1020/40 Intelligent SCSI Processor Driver (PCI) for PCI PowerMac
 * Written by Rob Clark (rclark@hmc.edu)
 * 
 * Credits: This driver's interface with the controller card is based 
 *          on qlogicisp.c written by Erik H. Moe (ehm@cris.com).  The 
 *          interface with the Open Firmware is based on mesh.c, written 
 *          by Paul Mackerras.  Special thanks to Michael AK Tesch for 
 *          his work on the ATI driver for linux-pmac, and his explaination
 *          of how to manually configure the PCI control register so that
 *          I could talk to the PCI scsi controller.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* Configuration section *****************************************************/
/* Set the following macro to 1 to reload the ISP1020's firmware.  This is   *
 * the latest firmware provided by QLogic.  This may be an earlier/later     *
 * revision than supplied by your board.                                     */
#define RELOAD_FIRMWARE		1

/* Set the following macro to 1 to reload the ISP1020's defaults from nvram. *
 * If you are not sure of your settings, leave this alone, the driver will   *
 * use a set of 'safe' defaults                                              */
#define USE_NVRAM_DEFAULTS	1

/* Controls macros used for debugging                                        */
#define DEBUG_ISP1020		0
#define DEBUG_INTERUPT          0
#define DEBUG_ENTERLEAVE        0
#define DEBUG_PARAMS            0
/* Which of the address properties to use for finding the address of the     *
 * memory space of our controller                                            */
#define ADDRESS                 2
/* End Configuration section *************************************************/


#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/blk.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/bios32.h>
#include <linux/sched.h>
#include <linux/unistd.h>
#include <asm/pgtable.h>
#include <asm/irq.h>
#include <asm/dbdma.h>
#include <asm/io.h>
#include <asm/prom.h>
#include <asm/system.h>
#include <asm/pci-bridge.h>

#include "scsi.h"
#include "hosts.h"
#include "sd.h"
#include "qlogic-pmac.h"
#include "qlogic-pmac-stuff.h"
#include "qlogic-pmac-fware.h"

struct proc_dir_entry proc_scsi_qlc = {
  PROC_SCSI_QLOGICISP, 7, "qlc", S_IFDIR | S_IRUGO | S_IXUGO, 2 };

struct Scsi_Host *irq2host[NR_IRQS];

/* Some Prototypes ************************************************************/
int         qlc_getboardtype( struct Scsi_Host * );
static void qlc_enable_irqs( struct Scsi_Host * );
static void qlc_disable_irqs( struct Scsi_Host * );
static int  qlc_mbox_command( struct Scsi_Host *, u_short [] );
static int  qlc_reset_hardware(struct Scsi_Host *);
static int  qlc_load_parameters(struct Scsi_Host *);
static int  qlc_set_defaults( struct Scsi_Host * );
#if USE_NVRAM_DEFAULTS
static int  qlc_get_defaults( struct Scsi_Host * );
static int  qlc_verify_nvram( struct Scsi_Host * );
u_short     qlc_read_nvram_word( struct Scsi_Host *, u_short );
#endif
static void qlc_intr_handler(int, void *, struct pt_regs *);
static int  qlc_return_status(struct Status_Entry *);
#if DEBUG_ISP1020
void        qlc_print_scsi_cmd( Scsi_Cmnd * );
void        qlc_print_status_entry( struct Status_Entry *status );
#endif
/* End Prototypes *************************************************************/

/******************************************************************************/
int
qlc_detect(Scsi_Host_Template *tp)
  {
  struct device_node  *qlcDevice;
  struct Scsi_Host    *host;
  struct qlc_hostdata *hostData;
  int    numQlcHosts=0;       /* Initially, no hosts have been detected */
  int    len;
  unsigned long addr;
  unsigned char bus, devfn;
  unsigned short cmd;

  ENTER("qlc_detect");
  
#if 0
  print_devices();            /* in arch/ppc/pmac/prom.c */   
#endif
  
  for( qlcDevice=find_devices("QLGC,ISP1020"); 
       qlcDevice!=0; 
       qlcDevice=qlcDevice->next )
    {
    DEBUG( printk( "qlc: found device \"%s\"\n", qlcDevice->name) );
    
    /********************************************************************
     * Map in the address of our device:                                */
    addr = (unsigned long)ioremap( qlcDevice->addrs[ADDRESS].address,
				   qlcDevice->addrs[ADDRESS].size );

    /********************************************************************
     * These next few lines |= the control register with a number, in   *
     * order to enable our device to responde to memory space access    */
    if (pci_device_loc(qlcDevice, &bus, &devfn) == 0) {
      pcibios_read_config_word(bus, devfn, PCI_COMMAND, &cmd);
      if (cmd != 0xffff) {
	DEBUG(printk("qlc: device config command reg: %x\n",cmd));
	cmd |= PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER
	  | PCI_COMMAND_SPECIAL | PCI_COMMAND_INVALIDATE;
	pcibios_write_config_word(bus, devfn, PCI_COMMAND, cmd);
      }
    }

    /********************************************************************/    
    /*** This is the part that actually registers the host adapter... ***/   
    host = scsi_register( tp, sizeof(struct qlc_hostdata) );
    hostData = (struct qlc_hostdata *)host->hostdata;
    if( host == 0 )
      panic("qlc: Couldn't register qlc host");

    /* This gives a unique id to all qlc hosts: */
    host->unique_id = numQlcHosts;
    host->this_id   = hostData->host_param.initiator_scsi_id; /* of the host */
    host->io_port       = addr;
    host->irq           = qlcDevice->intrs[0];
    
    /* the host->max_id and host->max_lun fields are 1+ the max number  *
     * of IDs or LUNs (narrow SCSI has 7 devices, so max_id = 8)        */
    switch( qlc_getboardtype(host) )
      {
      case NARROW:
	hostData->host_param.wide  = 0;
	hostData->host_param.ultra = 0;
	hostData->board_type       = "PCI Narrow";
	host->max_id      = 8;
	break;
      case WIDE:
	hostData->host_param.wide  = 1;
	hostData->host_param.ultra = 0;
	hostData->board_type       = "PCI Wide";
	host->max_id      = 16;
	break;
      case DIFFERENTIAL:
	hostData->host_param.wide  = 1;
	hostData->host_param.ultra = 0;
	hostData->board_type       = "PCI W/D";
	host->max_id      = 16;
	break;
      case ULTRA_NARROW:
	hostData->host_param.wide  = 0;
	hostData->host_param.ultra = 1;
	hostData->board_type       = "PCI Ultra";
	host->max_id      = 8;
	break;
      case ULTRA_WIDE:
	hostData->host_param.wide  = 1;
	hostData->host_param.ultra = 1;
	hostData->board_type       = "PCI Ultra-Wide";
	host->max_id      = 16;
	break;
      case ULTRA_DIFFERENTIAL:
	hostData->host_param.wide  = 1;
	hostData->host_param.ultra = 1;
	hostData->board_type       = "PCI Ultra-W/D";
	host->max_id      = 16;
	break;
      default:
	printk("qlc: could not identfy board type... assuming NARROW\n");
	hostData->host_param.wide  = 0;
	hostData->host_param.ultra = 0;
	hostData->board_type       = "Magical Mystery Board";
	host->max_id      = 8;
      }
    host->max_lun      = 8;
    host->max_channel  = 0;   /* This driver may work on multichannel   * 
			       * boards based on the ISP1020/ISP1040,   *
			       * but this will need to be changed       */
    hostData->slot     = get_property( qlcDevice,"AAPL,slot-name", &len );
    hostData->revision = *get_property( qlcDevice, "revision-id", &len );
    /***********************************************************/
    note_scsi_host(qlcDevice, host);    /* defined in setup.c  */
    
    if( request_irq(host->irq, qlc_intr_handler, SA_INTERRUPT,
		    "qlogicisp", NULL) )
      {
      printk("qlc: interrupt %d already in use\n", host->irq);
      scsi_unregister(host);
      continue;
      }
    if( check_region(host->io_port, qlcDevice->addrs[ADDRESS].size) )
      {
      printk("qlc: i/o region 0x%04x-0x%04x already in use\n",
	     host->io_port, host->io_port + 0xff);
      free_irq(host->irq, NULL);
      scsi_unregister(host);
      continue;
      }
    
    request_region( host->io_port, 0xff, "qlogicisp" );
    memset( irq2host, 0, sizeof(irq2host) );
    irq2host[host->irq] = host;
    
    if( qlc_reset_hardware(host) ||     /* If we can't reload the isp's firmware, bail out! */
        qlc_set_defaults(host)   ||     /* Set the defaults, or load 'em from NVRAM         */
        qlc_load_parameters(host) )     /* Load the parameters for the RISC (ISP)           */
      {
      scsi_unregister(host);
      return 0;
      }
    
    outw(0x0, host->io_port + PCI_SEMAPHORE);
    outw(HCCR_CLEAR_RISC_INTR, host->io_port + HOST_HCCR);
    
    qlc_enable_irqs(host);
    
    numQlcHosts++;                      /* Increment the number of detected hosts */
    }
  
  LEAVE("qlc_detect");
  return numQlcHosts;
  }

/******************************************************************************/
int 
qlc_getboardtype( struct Scsi_Host *host )
  {
  u_short param[6];
  unsigned long flags;
  save_flags(flags);
  
  param[0] = MBOX_READ_RAM_WORD;
  param[1] = BOARD_TYPE_ADDRESS;
  param[2] = 0;
  qlc_mbox_command(host, param);
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: error finding board type\n");
    return 1;
    }
  
  return (param[2] & BOARD_TYPE_MASK);
  }

/******************************************************************************/
int 
qlc_release( struct Scsi_Host *host )
  {
  ENTER("qlc_release");
  
  LEAVE("qlc_release");
  return 1;
  }

/******************************************************************************/
const char * 
qlc_info( struct Scsi_Host *host )
  {
  static char buf[80];
  struct qlc_hostdata *hostData;
  
  ENTER("qlc_info");
  
  hostData = (struct qlc_hostdata *) host->hostdata;
  sprintf(buf,
	  "QLogic ISP1020 %s SCSI for Macintosh rev. %d in PCI slot %s, irq %d, address 0x%x",
	  hostData->board_type, hostData->revision, hostData->slot, host->irq, host->io_port);
  
  LEAVE("qlc_info");
  
  return buf;
  }

/******************************************************************************/
int 
qlc_queue( Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *) )
  {
  u_int inPtr, outPtr;
  struct Command_Entry *cmdEntry;
  struct Scsi_Host *host;
  struct qlc_hostdata *hostData;
  int numFree=0;
  
  ENTER("qlc_queue");
  DEBUG(qlc_print_scsi_cmd(cmd));
  
  host = cmd->host;
  cmd->scsi_done = done;
  hostData = (struct qlc_hostdata *)host->hostdata;
  
  outPtr = inw( host->io_port + MBOX4 );
  inPtr  = hostData->req_in_ptr;
  
  DEBUG(printk("qlc: outPtr = %d; inPtr = %d\n", outPtr,inPtr));
  
  cmdEntry = (struct Command_Entry *)&hostData->req[inPtr][0];
  inPtr  = (inPtr +1) & QLOGICISP_REQ_QUEUE_LEN;
  
  if( inPtr == outPtr )
    {
    printk("qlc: request queue overflow\n");
    return 1;
    }
  
  if( hostData->send_marker )
    {
    struct Marker_Entry *markerEntry;
    
    DEBUG(printk("qlc: adding marker entry\n"));
    
    markerEntry = (struct Marker_Entry *)cmdEntry;
    memset( markerEntry, 0, sizeof(struct Marker_Entry) );
    
    markerEntry->hdr.entry_type = ENTRY_MARKER;
    markerEntry->hdr.entry_cnt  = 1;
    markerEntry->modifier = SYNC_ALL;
    
    hostData->send_marker = 0;
    
    if( ((inPtr+1) & QLOGICISP_REQ_QUEUE_LEN) == outPtr )
      {
      outw( inPtr, host->io_port + MBOX4 );
      hostData->req_in_ptr = inPtr;
      printk("qlc: request queue overflow\n");
      return 1;
      }
    
    cmdEntry = (struct Command_Entry *)&hostData->req[inPtr][0];
    inPtr = (inPtr+1) & QLOGICISP_REQ_QUEUE_LEN;
    }
  
  memset( cmdEntry, 0, sizeof(struct Command_Entry) );
  cmdEntry->hdr.entry_type = ENTRY_COMMAND;
  cmdEntry->hdr.entry_cnt  = 1;
  cmdEntry->handle         = __ER_long((u_int)cmd);
  cmdEntry->target_lun     = cmd->lun;
  cmdEntry->target_id      = cmd->target;
  cmdEntry->cdb_length     = __ER_short(cmd->cmd_len);
  cmdEntry->control_flags  = __ER_short(CFLAG_READ | CFLAG_WRITE);
  cmdEntry->time_out       = __ER_short(30);
  
  memcpy( cmdEntry->cdb, cmd->cmnd, cmd->cmd_len );
  
  if( cmd->use_sg )
    {
    struct Continuation_Entry *contEntry;
    int i;
    int sgCount,n;
    struct scatterlist *sg;

    DEBUG(printk("qlc: doing a sg\n"));

    sgCount = cmd->use_sg;
    cmdEntry->segment_cnt = __ER_short(sgCount);
    sg = (struct scatterlist *) cmd->request_buffer;
    
    /* fill in first four sg entries: */
    n = sgCount;
    if( n > 4 )
      n = 4;
    for( i=0; i<n; i++ )
      {
      cmdEntry->dataseg[i].d_base  = (u_int)__ER_long(virt_to_bus(sg->address));
      cmdEntry->dataseg[i].d_count = __ER_long(sg->length);
      ++sg;
      }
    sgCount -= 4;
    
    while( sgCount > 0 )
      {
      ++cmdEntry->hdr.entry_cnt;
      contEntry = (struct Continuation_Entry *)&hostData->req[inPtr][0];
      inPtr = (inPtr + 1) & QLOGICISP_REQ_QUEUE_LEN;
      if(inPtr == outPtr) 
	{
	printk("qlc: unexpected request queue overflow\n");
	return 1;
	}
      TRACE("queue continuation", in_ptr, 0);
      contEntry->hdr.entry_type = ENTRY_CONTINUATION;
      contEntry->hdr.entry_cnt  = 0;
      contEntry->hdr.sys_def_1  = 0;
      contEntry->hdr.flags      = 0;
      contEntry->reserved = __ER_long(0);
      n = sgCount;
      if( n > 7 )
	n = 7;
      for( i=0; i<n; ++i ) 
	{
	contEntry->dataseg[i].d_base = (u_int)__ER_long(virt_to_bus(sg->address));
	contEntry->dataseg[i].d_count = __ER_long(sg->length);
	++sg;
	}
      sgCount -= n;
      }
    }
  else
    {
    cmdEntry->dataseg[0].d_base  = __ER_long((u_int)virt_to_bus(cmd->request_buffer));
    cmdEntry->dataseg[0].d_count = __ER_long((u_int)cmd->request_bufflen);
    cmdEntry->segment_cnt = __ER_short(1);
    }
  
  outw( inPtr, host->io_port + MBOX4 );
  hostData->req_in_ptr = inPtr;
  
  numFree = QLOGICISP_REQ_QUEUE_LEN - REQ_QUEUE_DEPTH( inPtr, outPtr );
  host->can_queue    = host->host_busy + numFree;
  host->sg_tablesize = QLOGICISP_MAX_SG(numFree);
  
  LEAVE("qlc_queue");
  return 0;
  }

/******************************************************************************/
int 
qlc_abort( Scsi_Cmnd *cmd )
  {
  u_short param[6];
  struct Scsi_Host *host;
  struct qlc_hostdata *hostData;
  int return_status = SCSI_ABORT_SUCCESS;
  u_int cmdaddr = (u_int)(cmd);
  
  ENTER("qlc_abort");
  
  host = cmd->host;
  hostData = (struct qlc_hostdata *)host->hostdata;
  
  qlc_disable_irqs(host);
  
  param[0] = MBOX_ABORT;
  /* format of param[1] is 0000iiii0000llll where i is ID and l is lun bits */
  param[1] = (((u_short)cmd->target) << 8) | ((u_short)cmd->lun & 0x000f);
  param[2] = cmdaddr >> 16;
  param[3] = cmdaddr & 0xffff;
  
  qlc_mbox_command( host, param );
  
  if( param[0] != MBOX_COMMAND_COMPLETE )
    {
    printk("qlc: scsi abort failure: %x\n", param[0]);
    return_status = SCSI_ABORT_ERROR;
    }
  
  qlc_enable_irqs(host);
  
  LEAVE("qlc_abort");  
  return return_status;
  }

/******************************************************************************/
int 
qlc_reset( Scsi_Cmnd *cmd, unsigned int how )
  {
  u_short param[6];
  struct Scsi_Host *host;
  struct qlc_hostdata *hostData;
  int return_status = SCSI_RESET_SUCCESS;
  
  ENTER("qlc_reset");
  
  host = cmd->host;
  hostData = (struct qlc_hostdata *)host->hostdata;
  
  param[0] = MBOX_BUS_RESET;
  param[1] = hostData->host_param.bus_reset_delay;
  
  qlc_disable_irqs(host);
  
  qlc_mbox_command(host, param);
  
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    printk("qlc: scsi bus reset failure: %x\n", param[0]);
    return_status = SCSI_RESET_ERROR;
    }
  
  qlc_enable_irqs(host);
  
  LEAVE("qlc_reset");
  return return_status;;
  }

/******************************************************************************/
int 
qlc_bios_param( Disk *disk, kdev_t n, int ip[] )
  {
  ENTER("qlc_bios_param");
  LEAVE("qlc_bios_param");
  return 1;
  }

/******************************************************************************/
static void 
qlc_enable_irqs( struct Scsi_Host *host )
  {
  ENTER("qlc: qlc_enable_irqs");
  outw(ISP_EN_INT|ISP_EN_RISC, host->io_port + PCI_INTF_CTL);
  LEAVE("qlc: qlc_enable_irqs");
  }

/******************************************************************************/
static void 
qlc_disable_irqs( struct Scsi_Host *host )
  {
  ENTER("qlc: qlc_disable_irqs");
  outw(0x0, host->io_port + PCI_INTF_CTL);
  LEAVE("qlc: qlc_disable_irqs");
  }

/******************************************************************************/
/*
 * currently, this is only called during initialization or abort/reset,
 * at which times interrupts are disabled, so polling is OK, I guess...
 */
static int 
qlc_mbox_command( struct Scsi_Host *host, u_short param[] )
  {
  int loop_count;
  
  if( mbox_param[param[0]] == 0 )
    return 1;
  
  loop_count = DEFAULT_LOOP_COUNT;
  while( --loop_count && inw(host->io_port + HOST_HCCR) & HCCR_HOST_INT )
    barrier();
  if( !loop_count )
    printk("qlc: mbox_command loop timeout #1\n");
  
  switch( mbox_param[param[0]] >> 4 ) 
    {
    case 6: outw(param[5], host->io_port + MBOX5);
    case 5: outw(param[4], host->io_port + MBOX4);
    case 4: outw(param[3], host->io_port + MBOX3);
    case 3: outw(param[2], host->io_port + MBOX2);
    case 2: outw(param[1], host->io_port + MBOX1);
    case 1: outw(param[0], host->io_port + MBOX0);
    }
  
  outw( 0x0, host->io_port + PCI_SEMAPHORE );
  outw( HCCR_CLEAR_RISC_INTR, host->io_port + HOST_HCCR );
  outw( HCCR_SET_HOST_INTR, host->io_port + HOST_HCCR );
  
  loop_count = DEFAULT_LOOP_COUNT;
  while( --loop_count && !(inw(host->io_port + PCI_INTF_STS) & 0x04) )
    barrier();
  if( !loop_count )
    printk("qlc: mbox_command loop timeout #2\n");
  
  loop_count = DEFAULT_LOOP_COUNT;
  while( --loop_count && inw(host->io_port + MBOX0) == 0x04 )
    barrier();
  if( !loop_count )
    printk("qlogicisp: mbox_command loop timeout #3\n");
  
  switch( mbox_param[param[0]] & 0xf ) 
    {
    case 6: param[5] = inw(host->io_port + MBOX5);
    case 5: param[4] = inw(host->io_port + MBOX4);
    case 4: param[3] = inw(host->io_port + MBOX3);
    case 3: param[2] = inw(host->io_port + MBOX2);
    case 2: param[1] = inw(host->io_port + MBOX1);
    case 1: param[0] = inw(host->io_port + MBOX0);
    }
  
  outw( 0x0, host->io_port + PCI_SEMAPHORE );
  outw( HCCR_CLEAR_RISC_INTR, host->io_port + HOST_HCCR );
  
  return 0;
  }

/******************************************************************************/
static int
qlc_reset_hardware( struct Scsi_Host *host )
  {
  u_short param[6];
  int loop_count;
  int i;
  
  ENTER("qlc_reset_hardware");
  
  outw( ISP_RESET, host->io_port + PCI_INTF_CTL );
  outw( HCCR_RESET, host->io_port + HOST_HCCR );
  outw( HCCR_RELEASE, host->io_port + HOST_HCCR );
  outw( HCCR_BIOS_DISABLE, host->io_port + HOST_HCCR );
  
  loop_count = DEFAULT_LOOP_COUNT;
  while( --loop_count && inw(host->io_port + HOST_HCCR) == RISC_BUSY )
    barrier();
  if( !loop_count )
    printk("qlc: reset_hardware loop timeout\n");
  
  outw( 0, host->io_port + ISP_CFG1 );

#if RELOAD_FIRMWARE
  for( i=0; i<risc_code_length01; i++ ) 
    {
    param[0] = MBOX_WRITE_RAM_WORD;
    param[1] = risc_code_addr01 + i;
    param[2] = risc_code01[i];
    qlc_mbox_command(host, param);
    
    if (param[0] != MBOX_COMMAND_COMPLETE) 
      {
      printk("qlc: firmware load failure\n");
      return 1;
      }
    }
#endif /* RELOAD_FIRMWARE */

  param[0] = MBOX_VERIFY_CHECKSUM;
  param[1] = risc_code_addr01;
  qlc_mbox_command( host,param );
  
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    printk("qlc: ram checksum failure\n");
    return 1;
    }
  
  param[0] = MBOX_EXEC_FIRMWARE;
  param[1] = risc_code_addr01;
  qlc_mbox_command( host,param );
  
  param[0] = MBOX_ABOUT_FIRMWARE;
  qlc_mbox_command( host,param );

  if( param[0] != MBOX_COMMAND_COMPLETE )
    {
    printk("qlc: about firmware failure\n");
    return 1;
    }
  
  DEBUG(printk("qlc: firmware major revision %d\n", param[1]));
  DEBUG(printk("qlc: firmware minor revision %d\n", param[2]));
  
  LEAVE("qlc_reset_hardware");
  
  return 0;
  }

/******************************************************************************/
static int
qlc_load_parameters( struct Scsi_Host *host )
  {
  int i, k;
  u_int queue_addr;
  u_short param[6];
  /*  u_short isp_cfg1; */
  unsigned long flags;
  struct qlc_hostdata *hostData =
    (struct qlc_hostdata *) host->hostdata;
  
  ENTER("qlc_load_parameters");
  
  save_flags(flags);
  cli();
  
#if 0
  /* Pause the risc, so we can talk to the flash */
  outw(HCCR_PAUSE, host->io_port + HOST_HCCR);
  while (!(inw(host->io_port+HOST_HCCR) & HCCR_PAUSED))
    ;
  /* contents of MEM_TIMING_REG |= 0xff00 */
  outw(((inw(host->io_port+MEM_TIMING_REG))|0xff00), host->io_port + MEM_TIMING_REG);
  /* contenets of MEM_BOUNDRY_REG = 8 */
  outw(8,host->io_port + MEM_BOUNDRY_REG);
  /* Unpause the risc */
  outw(HCCR_RELEASE, host->io_port + HOST_HCCR);
  while ((inw(host->io_port+HOST_HCCR) & HCCR_PAUSED))
    ;
#endif
  
  /* Setup PCI control parameters */
  param[0] = MBOX_SET_PCI_CONTROL_PARAMS;
  param[1] = hostData->host_param.data_dma_burst_enable << 1;
  param[2] = hostData->host_param.command_dma_burst_enable << 1;
  qlc_mbox_command(host, param);
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: set pci control parameter failure\n");
    return 1;
    }
  
  /* If either DMA burst flag is enabled, enable bursting */  
  if( hostData->host_param.data_dma_burst_enable ||
       hostData->host_param.command_dma_burst_enable )
    outw( (hostData->host_param.fifo_threshold | BURST_ENABLE), host->io_port + ISP_CFG1 );
  else
    outw( hostData->host_param.fifo_threshold, host->io_port + ISP_CFG1 );
  
  /* Set the termination... for now, it is hardcoded for autotermination */
  param[0] = MBOX_WRITE_RAM_WORD;
  param[1] = TERMINATION_ADDRESS;
  param[2] = 0;  
  qlc_mbox_command( host,param );
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: set termination parameter failure\n");
    return 1;
    }

  /* Set the controllers ID... (normally 7) */
  param[0] = MBOX_SET_INIT_SCSI_ID;
  param[1] = hostData->host_param.initiator_scsi_id;
  qlc_mbox_command(host, param);
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: set initiator id failure\n");
    return 1;
    }
  
  /* Set the selection timout */
  param[0] = MBOX_SET_SELECT_TIMEOUT;
  param[1] = hostData->host_param.selection_timeout;
  qlc_mbox_command(host, param);  
  if (param[0] != MBOX_COMMAND_COMPLETE) 
    {
    restore_flags(flags);
    printk("qlc: set selection timeout failure\n");
    return 1;
    }
  
  /* Set retry count */
  param[0] = MBOX_SET_RETRY_COUNT;
  param[1] = hostData->host_param.retry_count;
  param[2] = hostData->host_param.retry_delay;
  qlc_mbox_command(host, param);
  if (param[0] != MBOX_COMMAND_COMPLETE) 
    {
    restore_flags(flags);
    printk("qlc: set retry count failure\n");
    return 1;
    }
  
  /* Set tag age limit */
  param[0] = MBOX_SET_TAG_AGE_LIMIT;
  param[1] = hostData->host_param.tag_aging;
  qlc_mbox_command(host, param);
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: set tag age limit failure\n");
    return 1;
    }
  
  /* Set Clock... */
  param[0] = MBOX_SET_CLOCK_RATE;
  param[1] = (hostData->host_param.ultra)? 60 : 40;
  qlc_mbox_command( host,param );
  if (param[0] != MBOX_COMMAND_COMPLETE) 
    {
    restore_flags(flags);
    printk("qlc: clock speed set failure\n");
    return 1;
    }
  
  /* Set active negotiation state */
  param[0] = MBOX_SET_ACTIVE_NEG_STATE;
#if 1  /* The way it was in the PC linux driver ??? */
  param[1] = (hostData->host_param.req_ack_active_negation << 4) |
             (hostData->host_param.data_line_active_negation << 5);
#else  /* The way it was in the mac driver ??? */
  param[1] = (hostData->host_param.req_ack_active_negation << 5) | 
             (hostData->host_param.data_line_active_negation << 4);  
#endif
  qlc_mbox_command( host,param );
  if( param[0] != MBOX_COMMAND_COMPLETE ) 
    {
    restore_flags(flags);
    printk("qlc: set active negation state failure\n");
    return 1;
    }
   
  /* Set async data setup time */
  param[0] = MBOX_SET_ASYNC_DATA_SETUP_TIME;
  param[1] = hostData->host_param.async_data_setup_time;
  qlc_mbox_command(host, param);
  if (param[0] != MBOX_COMMAND_COMPLETE) 
    {
    restore_flags(flags);
    printk("qlc: async data setup time failure\n");
    return 1;
    }
  
  for( i=0; i<MAX_TARGETS; i++ )
    {
    if( !hostData->dev_param[i].device_enable )
      continue;
    
    if( hostData->host_param.wide ||
	hostData->host_param.ultra )
      hostData->dev_param[i].device_flags |= CAP_RENEGOTIATE_ON_ERROR;
    
    param[0] = MBOX_SET_TARGET_PARAMS;
    param[1] = i << 8;
    param[2] = hostData->dev_param[i].device_flags << 8;
    param[3] = (hostData->dev_param[i].synchronous_offset << 8) | 
               hostData->dev_param[i].synchronous_period;
    
    qlc_mbox_command(host, param);
    
    if( param[0] != MBOX_COMMAND_COMPLETE )
      {
      restore_flags(flags);
      printk("qlc: set target parameter failure\n");
      return 1;
      }
    
    for( k=0; k<MAX_LUNS; k++) 
      {
      param[0] = MBOX_SET_DEV_QUEUE_PARAMS;
      param[1] = (i << 8) | k;
      param[2] = hostData->host_param.max_queue_depth;
      param[3] = hostData->dev_param[i].execution_throttle;
      
      qlc_mbox_command(host, param);
      
      if (param[0] != MBOX_COMMAND_COMPLETE) 
	{
	restore_flags(flags);
	printk("qlc: set device queue parameter failure\n");
	return 1;
	}
      }
    }
  
  queue_addr = (u_int) virt_to_bus(&hostData->res[0][0]);
  
  param[0] = MBOX_INIT_RES_QUEUE;
  param[1] = RES_QUEUE_LEN + 1;
  param[2] = (u_short) (queue_addr >> 16);
  param[3] = (u_short) (queue_addr & 0xffff);
  param[4] = 0;
  param[5] = 0;
  
  qlc_mbox_command(host, param);
  
  if( param[0] != MBOX_COMMAND_COMPLETE )
    {
    restore_flags(flags);
    printk("qlc: set response queue failure\n");
    return 1;
    }
  
  queue_addr = (u_int) virt_to_bus(&hostData->req[0][0]);
  
  param[0] = MBOX_INIT_REQ_QUEUE;
  param[1] = QLOGICISP_REQ_QUEUE_LEN + 1;
  param[2] = (u_short) (queue_addr >> 16);
  param[3] = (u_short) (queue_addr & 0xffff);
  param[4] = 0;
  
  qlc_mbox_command(host, param);
  
  if (param[0] != MBOX_COMMAND_COMPLETE) 
    {
    restore_flags(flags);
    printk("qlc: set request queue failure\n");
    return 1;
    }
  
  restore_flags(flags);
  
  LEAVE("qlc_load_parameters");
  
  return 0;
  }

/******************************************************************************/
static int 
qlc_set_defaults( struct Scsi_Host *host )
  {
  struct qlc_hostdata *hostData;
  int i;
#if USE_NVRAM_DEFAULTS
  /* If qlc_get_defaults was successful, return.  Otherwise set defaults. */
  if( qlc_get_defaults(host) == 0 )
    return 0;
#endif  
  hostData = (struct qlc_hostdata *) host->hostdata;
  
  ENTER("qlc_set_defaults");
  
  hostData->host_param.fifo_threshold = 3;
  hostData->host_param.host_adapter_enable = 1;
  hostData->host_param.initiator_scsi_id = 7;
  hostData->host_param.bus_reset_delay = 3;
  hostData->host_param.retry_count = 0;
  hostData->host_param.retry_delay = 5;
  hostData->host_param.async_data_setup_time = 6;
  hostData->host_param.req_ack_active_negation = 1;
  hostData->host_param.data_line_active_negation = 1;
  hostData->host_param.data_dma_burst_enable = 1;
  hostData->host_param.command_dma_burst_enable = 1;
  hostData->host_param.tag_aging = 8;
  hostData->host_param.selection_timeout = 250;
  hostData->host_param.max_queue_depth = 400;
  
  for( i=0; i<MAX_TARGETS; i++ ) 
    {
    hostData->dev_param[i].device_flags = ( 0
					    | CAP_RENEGOTIATE_ON_ERROR
					    /*  | CAP_STOP_QUEUE_ON_CHECK */
					    | CAP_AUTO_REQUEST_SENSE /* ALWAYS do this */
					    | CAP_TAGGED_QUEUING
					    | CAP_SYNC_DATA_TRANSFERS
					    | CAP_PARITY_CHECKING
					    | CAP_DISCONNECT_PRIV
      );
    if( hostData->host_param.wide )
      hostData->dev_param[i].device_flags |=  CAP_WIDE_DATA_TRANSFERS;

    if( hostData->host_param.ultra )
      hostData->dev_param[i].synchronous_period = 12;  /*  50 ns */
    else
      hostData->dev_param[i].synchronous_period = 25;  /* 100 ns */
    hostData->dev_param[i].synchronous_offset = 12;
    hostData->dev_param[i].execution_throttle = 16;
    hostData->dev_param[i].device_enable = 1;
    }
  
  LEAVE("qlc_set_defaults");
  
  return 0;
  }

/******************************************************************************/
#if USE_NVRAM_DEFAULTS
static int 
qlc_get_defaults( struct Scsi_Host *host )
  {
  int i;
  u_short value;
  struct qlc_hostdata *hostData =
    (struct qlc_hostdata *) host->hostdata;
  
  ENTER("qlc_get_defaults");
  
#if 0
  if( !qlc_verify_nvram(host) )
    {
    printk("qlc: nvram checksum failure\n");
    printk("qlc: attempting to use default parameters\n");
    return 1;
    }
#endif
  
  value = qlc_read_nvram_word(host, 2);
  hostData->host_param.fifo_threshold = (value >> 8) & 0x03;
  hostData->host_param.host_adapter_enable = (value >> 11) & 0x01;
  hostData->host_param.initiator_scsi_id = (value >> 12) & 0x0f;
  
  value = qlc_read_nvram_word(host, 3);
  hostData->host_param.bus_reset_delay = value & 0xff;
  hostData->host_param.retry_count = value >> 8;
  
  value = qlc_read_nvram_word(host, 4);
  hostData->host_param.retry_delay = value & 0xff;
  hostData->host_param.async_data_setup_time = (value >> 8) & 0x0f;
  hostData->host_param.req_ack_active_negation = (value >> 12) & 0x01;
  hostData->host_param.data_line_active_negation = (value >> 13) & 0x01;
  hostData->host_param.data_dma_burst_enable = (value >> 14) & 0x01;
  hostData->host_param.command_dma_burst_enable = (value >> 15);
  
  value = qlc_read_nvram_word(host, 5);
  hostData->host_param.tag_aging = value & 0xff;
  
  value = qlc_read_nvram_word(host, 6);
  hostData->host_param.selection_timeout = value & 0xffff;
  
  value = qlc_read_nvram_word(host, 7);
  hostData->host_param.max_queue_depth = value & 0xffff;
  
#if DEBUG_PARAMS
  printk("qlc: fifo threshold=%d\n",
	 hostData->host_param.fifo_threshold);
  printk("qlc: initiator scsi id=%d\n",
	 hostData->host_param.initiator_scsi_id);
  printk("qlc: bus reset delay=%d\n",
	 hostData->host_param.bus_reset_delay);
  printk("qlc: retry count=%d\n",
	 hostData->host_param.retry_count);
  printk("qlc: retry delay=%d\n",
	 hostData->host_param.retry_delay);
  printk("qlc: async data setup time=%d\n",
	 hostData->host_param.async_data_setup_time);
  printk("qlc: req/ack active negation=%d\n",
	 hostData->host_param.req_ack_active_negation);
  printk("qlc: data line active negation=%d\n",
	 hostData->host_param.data_line_active_negation);
  printk("qlc: data DMA burst enable=%d\n",
	 hostData->host_param.data_dma_burst_enable);
  printk("qlc: command DMA burst enable=%d\n",
	 hostData->host_param.command_dma_burst_enable);
  printk("qlc: tag age limit=%d\n",
	 hostData->host_param.tag_aging);
  printk("qlc: selection timeout limit=%d\n",
	 hostData->host_param.selection_timeout);
  printk("qlc: max queue depth=%d\n",
	 hostData->host_param.max_queue_depth);
#endif /* DEBUG_PARAMS */
  
  for( i=0; i<MAX_TARGETS; i++ )
    {
    value = qlc_read_nvram_word(host, 14 + i * 3);
    hostData->dev_param[i].device_flags = value & 0xff;
    hostData->dev_param[i].execution_throttle = value >> 8;
    
    value = qlc_read_nvram_word(host, 15 + i * 3);
    hostData->dev_param[i].synchronous_period = value & 0xff;
    hostData->dev_param[i].synchronous_offset = (value >> 8) & 0x0f;
    hostData->dev_param[i].device_enable = (value >> 12) & 0x01;
    
#if DEBUG_PARAMS
    printk("qlc: target 0x%02x\n", i);
    printk("qlc:     device flags=0x%02x\n",
	   hostData->dev_param[i].device_flags);
    printk("qlc:     execution throttle=%d\n",
	   hostData->dev_param[i].execution_throttle);
    printk("qlc:     synchronous period=%d\n",
	   hostData->dev_param[i].synchronous_period);
    printk("qlc:     synchronous offset=%d\n",
	   hostData->dev_param[i].synchronous_offset);
    printk("qlc:     device enable=%d\n",
	   hostData->dev_param[i].device_enable);
#endif /* DEBUG_PARAMS */
    }
  
  LEAVE("qlc_get_defaults");
  
  return 0;
  }


/******************************************************************************/
#define ISP1020_NVRAM_LEN	0x40
#define ISP1020_NVRAM_SIG1	0x5349
#define ISP1020_NVRAM_SIG2	0x2050
static int 
qlc_verify_nvram( struct Scsi_Host *host )
  {
  int	i;
  u_short value;
  u_char checksum = 0;
  
  for( i=0; i<ISP1020_NVRAM_LEN; i++ )
    {
    value = qlc_read_nvram_word(host, i);
    
    switch(i)
      {
      case 0:
	if( value != ISP1020_NVRAM_SIG1 )
	  return 0;
	break;
      case 1:
	if( value != ISP1020_NVRAM_SIG2 )
	  return 0;
	break;
      case 2:
	if( (value & 0xff) != 0x02) 
	  return 0;
	break;
      }
    checksum += value & 0xff;
    checksum += value >> 8;
    }
  
  return (checksum == 0);
  }

#define NVRAM_DELAY() udelay(2) /* 2 microsecond delay */

/******************************************************************************/
u_short 
qlc_read_nvram_word( struct Scsi_Host *host, u_short byte )
  {
  int i;
  u_short value, output, input;
  
  byte &= 0x3f; byte |= 0x0180;
  
  for( i=8; i>=0; i-- )
    {
    output = ((byte >> i) & 0x1) ? 0x4 : 0x0;
    outw(output | 0x2, host->io_port + PCI_NVRAM); NVRAM_DELAY();
    outw(output | 0x3, host->io_port + PCI_NVRAM); NVRAM_DELAY();
    outw(output | 0x2, host->io_port + PCI_NVRAM); NVRAM_DELAY();
    }
  
  for( i=0xf, value=0; i>=0; i-- )
    {
    value <<= 1;
    outw(0x3, host->io_port + PCI_NVRAM); NVRAM_DELAY();
    input = inw(host->io_port + PCI_NVRAM); NVRAM_DELAY();
    outw(0x2, host->io_port + PCI_NVRAM); NVRAM_DELAY();
    if (input & 0x8) value |= 1;
    }
  outw(0x0, host->io_port + PCI_NVRAM); NVRAM_DELAY();
  
  return value;
  }
#endif /* USE_NVRAM_DEFAULTS */

/******************************************************************************/
void 
qlc_intr_handler( int irq, void *dev_id, struct pt_regs *regs )
  {
  Scsi_Cmnd *cmd;
  struct Status_Entry *sts;
  struct Scsi_Host *host;
  struct qlc_hostdata *hostData;
  u_int inPtr, outPtr;
  u_short status;
  
  ENTER("qlc_intr_handler");
  
  host = irq2host[irq];
  if( !host ) 
    {
    printk("qlc: unexpected interrupt on line %d\n", irq);
    return;
    }
  hostData = (struct qlc_hostdata *) host->hostdata;
  
  DEBUG(printk("qlc: interrupt on line %d\n", irq));
  
  if( !(inw(host->io_port + PCI_INTF_STS) & 0x04) ) 
    {
    /* spurious interrupts can happen legally */
    DEBUG(printk("qlc: got spurious interrupt\n"));
    return;
    }
  inPtr = inw(host->io_port + MBOX5);
  outw(HCCR_CLEAR_RISC_INTR, host->io_port + HOST_HCCR);
  
  if( (inw(host->io_port + PCI_SEMAPHORE) & ASYNC_EVENT_INTERRUPT) ) 
    {
    status = inw(host->io_port + MBOX0);
    
    DEBUG(printk("qlc: mbox completion status: %x\n",
		      status));
    
    switch( status ) 
      {
      case ASYNC_SCSI_BUS_RESET:
      case EXECUTION_TIMEOUT_RESET:
	hostData->send_marker = 1;
	break;
      case INVALID_COMMAND:
      case HOST_INTERFACE_ERROR:
      case COMMAND_ERROR:
      case COMMAND_PARAM_ERROR:
	printk("qlc: bad mailbox return status\n");
	break;
      }
    outw(0x0, host->io_port + PCI_SEMAPHORE);
    }
  outPtr = hostData->res_out_ptr;
  
  DEBUG_INT(printk("qlc: response queue update\n"));
  DEBUG_INT(printk("qlc: response queue depth %d\n",
		   REQ_QUEUE_DEPTH(inPtr, outPtr)));
  
  DEBUG_INT(printk("qlc: "));
  while( outPtr != inPtr ) 
    {
    DEBUG_INT(printk("."));
    sts = (struct Status_Entry *) &hostData->res[outPtr][0];
    outPtr = (outPtr + 1) & RES_QUEUE_LEN;
    
    cmd = (Scsi_Cmnd *)(__ER_long(sts->handle));
    TRACE("done", outPtr, cmd);
    
    status = __ER_short(sts->completion_status);
    if( (status == CS_RESET_OCCURRED) ||
	(status == CS_ABORTED) ||
	(__ER_short(sts->status_flags) & STF_BUS_RESET))
      hostData->send_marker = 1;
    
    if( __ER_short(sts->state_flags) & SF_GOT_SENSE )
      memcpy(cmd->sense_buffer, sts->req_sense_data,sizeof(cmd->sense_buffer));
    
    DEBUG_INT(qlc_print_status_entry(sts));
    
    /* entry_type is a char, so we don't have to worry about it's endianess */
    if( sts->hdr.entry_type == ENTRY_STATUS )
      cmd->result = qlc_return_status(sts);
    else
      cmd->result = DID_ERROR << 16;
    
    outw(outPtr, host->io_port + MBOX5);
    (*cmd->scsi_done)(cmd);
    }
  DEBUG_INT(printk("\n"));
  hostData->res_out_ptr = outPtr;
  
  LEAVE("qlc_intr_handler");
  }

/******************************************************************************/
static int
qlc_return_status( struct Status_Entry *sts )
  {
  int host_status = DID_ERROR;
#if DEBUG_ISP1020
  static char *reason[] = {
    "DID_OK",
    "DID_NO_CONNECT",
    "DID_BUS_BUSY",
    "DID_TIME_OUT",
    "DID_BAD_TARGET",
    "DID_ABORT",
    "DID_PARITY",
    "DID_ERROR",
    "DID_RESET",
    "DID_BAD_INTR"
  };
#endif /* DEBUG_ISP1020 */
  ENTER("qlc_return_status");
  
  DEBUG_INT(printk("qlc: completion status = 0x%04x\n",__ER_short(sts->completion_status)));
  
  switch(__ER_short(sts->completion_status)) 
    {
    case CS_COMPLETE:
      host_status = DID_OK;
      break;
    case CS_INCOMPLETE:
      if(!(__ER_short(sts->state_flags) & SF_GOT_BUS))
	host_status = DID_NO_CONNECT;
      else if(!(__ER_short(sts->state_flags) & SF_GOT_TARGET))
	host_status = DID_BAD_TARGET;
      else if(!(__ER_short(sts->state_flags) & SF_SENT_CDB))
	host_status = DID_ERROR;
      else if(!(__ER_short(sts->state_flags) & SF_TRANSFERRED_DATA))
	host_status = DID_ERROR;
      else if(!(__ER_short(sts->state_flags) & SF_GOT_STATUS))
	host_status = DID_ERROR;
      else if(!(__ER_short(sts->state_flags) & SF_GOT_SENSE))
	host_status = DID_ERROR;
      break;
    case CS_DMA_ERROR:
    case CS_TRANSPORT_ERROR:
      host_status = DID_ERROR;
      break;
    case CS_RESET_OCCURRED:
      host_status = DID_RESET;
      break;
    case CS_ABORTED:
      host_status = DID_ABORT;
      break;
    case CS_TIMEOUT:
      host_status = DID_TIME_OUT;
      break;
    case CS_DATA_OVERRUN:
    case CS_COMMAND_OVERRUN:
    case CS_STATUS_OVERRUN:
    case CS_BAD_MESSAGE:
    case CS_NO_MESSAGE_OUT:
    case CS_EXT_ID_FAILED:
    case CS_IDE_MSG_FAILED:
    case CS_ABORT_MSG_FAILED:
    case CS_NOP_MSG_FAILED:
    case CS_PARITY_ERROR_MSG_FAILED:
    case CS_DEVICE_RESET_MSG_FAILED:
    case CS_ID_MSG_FAILED:
    case CS_UNEXP_BUS_FREE:
      host_status = DID_ERROR;
      break;
    case CS_DATA_UNDERRUN:
      host_status = DID_OK;
      break;
    default:
      printk("qlc: unknown completion status 0x%04x\n",__ER_short(sts->completion_status));
      host_status = DID_ERROR;
      return DID_ERROR;    /* This isn't right, but it will screw up the mid-level driver 
			    * enough to not hang the machine during boot ??? */
      break;
    }
  
  DEBUG(printk("qlc: host status (%s) scsi status %x\n",
	       reason[host_status], __ER_short(sts->scsi_status)));
  LEAVE("qlc_return_status");
  return (__ER_short(sts->scsi_status) & STATUS_MASK) | (host_status << 16);
  }

/******************************************************************************/
#if DEBUG_ISP1020
void 
qlc_print_scsi_cmd( Scsi_Cmnd *cmd )
  {
  int i;
  
  printk("qlc: target = 0x%02x, lun = 0x%02x, cmd_len = 0x%02x\n",
	 cmd->target, cmd->lun, cmd->cmd_len);
  printk("qlc: command = ");
  for (i = 0; i < cmd->cmd_len; i++)
    printk("0x%02x ", cmd->cmnd[i]);
  printk("\n");
  }
/******************************************************************************/
void 
qlc_print_status_entry( struct Status_Entry *status )
  {
  int i;
  
  printk("qlc: entry count = 0x%02x, type = 0x%02x, flags = 0x%02x\n",
	 status->hdr.entry_cnt, status->hdr.entry_type, status->hdr.flags);
  printk("qlc: scsi status = 0x%04x, completion status = 0x%04x\n",
	 status->scsi_status, status->completion_status);
  printk("qlc: state flags = 0x%04x, status flags = 0x%04x\n",
	 status->state_flags, status->status_flags);
  printk("qlc: time = 0x%04x, request sense length = 0x%04x\n",
	 status->time, status->req_sense_len);
  printk("qlc: residual transfer length = 0x%08x\n", status->residual);
/*  
 *  for( i=0; i<status->req_sense_len; i++ )
 *    printk("qlc: sense data = 0x%02x\n", status->req_sense_data[i]);
 */
  }
#endif /* DEBUG_ISP1020 */

#ifdef MODULE
static Scsi_Host_Template driver_template = SCSI_QLOGIC_PMAC;

#include "scsi_module.c"
#endif
